# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from sqlalchemy.dialects import registry

registry.register('kylin', 'kylinpy.sqla_dialect', 'KylinDialect')

pytest_plugins = [
    'tests.fixtures.api',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.logger import logger


def private_v4_api_warnings(fn):
    def wrapper(*args, **kwargs):
        logger.warning('Use internal v4-API call, this API-CALL is not in KE4 official manual')
        return fn(*args, **kwargs)
    return wrapper



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pytest

from kylinpy.client import HTTPError
from kylinpy.kylinpy import create_kylin
from kylinpy.exceptions import KylinQueryError
from .test_client import MockException


class TestKE4Service(object):
    @property
    def cluster(self):
        return create_kylin('kylin://ADMIN:KYLIN@example?version=v4')

    @property
    def project(self):
        return create_kylin('kylin://ADMIN:KYLIN@example/ssb?version=v4')

    def test_projects(self, v4_api):
        rv = self.project.service.projects(headers={})
        assert [e['name'] for e in rv] == ['ssb']

    def test_jobs(self, v4_api):
        rv = self.project.service.jobs(headers={}, params={'time_filter': 4})
        assert [e['project'] for e in rv] == ['test']

    def test_tables_and_columns(self, v4_api):
        rv = self.project.service.tables_and_columns(headers={})
        assert sorted(list(rv.keys())) == [
            'DEFAULT.KYLIN_ACCOUNT',
            'DEFAULT.KYLIN_CAL_DT',
            'DEFAULT.KYLIN_CATEGORY_GROUPINGS',
            'DEFAULT.KYLIN_COUNTRY',
            'DEFAULT.KYLIN_SALES',
        ]

    def test_tables_in_hive(self, v4_api):
        rv = self.project.service.tables_in_hive(headers={})
        assert sorted(list(rv.keys())) == [
            'SSB.CUSTOMER',
            'SSB.DATES',
            'SSB.LINEORDER',
            'SSB.PART',
            'SSB.P_LINEORDER',
            'SSB.SUPPLIER',
        ]

    def test_query(self, v4_api):
        rv = self.project.service.query(sql='select count(*) from P_LINEORDER', headers={})
        assert 'columnMetas' in rv
        assert 'results' in rv

    def test_error_query(self, mocker):
        mocker.patch('kylinpy.service.KE4Service.api.query', return_value={'exceptionMessage': 'foobar'})

        with pytest.raises(KylinQueryError):
            self.project.service.query(sql='select count(*) from P_LINEORDER')

    def test_http_error_query(self, mocker):
        mc = mocker.patch('kylinpy.client.client.Client._make_request')
        mc.side_effect = MockException(500)
        with pytest.raises(HTTPError):
            self.project.service.query(sql='select count(*) from P_LINEORDER')

    def test_get_authentication(self, v4_api):
        rv = self.project.service.get_authentication(headers={})
        assert 'username' in rv
        assert 'authorities' in rv

    def test_models(self, v4_api):
        rv = self.project.service.models()
        assert 'kylin_sales_model' == rv[0]['alias']
        assert 'kylin_sales_model' == rv[0]['name']
        assert 'lookups' in rv[0]
        assert 'join_tables' in rv[0]

    def test_model_desc(self, v4_api):
        rv = self.project.service.model_desc('kylin_sales_model')
        assert rv['name'] == 'kylin_sales_model'
        assert 'measures' in rv
        assert 'dimensions' in rv
        assert 'aggregation_groups' in rv



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .client import Client  # noqa
from .exceptions import (  # noqa
    HTTPError,
    BadRequestsError,
    UnauthorizedError,
    ForbiddenError,
    NotFoundError,
    MethodNotAllowedError,
    PayloadTooLargeError,
    UnsupportedMediaTypeError,
    TooManyRequestsError,
    InternalServerError,
    ServiceUnavailableError,
    GatewayTimeoutError,
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re

from sqlalchemy.types import (
    BIGINT,
    BOOLEAN,
    CHAR,
    DATE,
    DATETIME,
    DECIMAL,
    FLOAT,
    INTEGER,
    SMALLINT,
    TIMESTAMP,
    VARCHAR,
)

KylinType = dict(
    CHAR=CHAR,
    STRING=VARCHAR,
    VARCHAR=VARCHAR,
    DECIMAL=DECIMAL,
    DOUBLE=FLOAT,
    FLOAT=FLOAT,
    BIGINT=BIGINT,
    LONG=BIGINT,
    INTEGER=INTEGER,
    INT=INTEGER,
    TINYINT=SMALLINT,
    SMALLINT=SMALLINT,
    INT4=BIGINT,
    LONG8=BIGINT,
    BOOLEAN=BOOLEAN,
    DATE=DATE,
    DATETIME=DATETIME,
    TIMESTAMP=TIMESTAMP,
)


def kylin_to_sqla(s):
    # the '|' operator is never greedy, so sorted keys by key length
    keys = list(sorted(KylinType.keys(), key=len, reverse=True))
    type_re = re.compile(
        r'^({})\(?(\d+)?,?(\d+)?\)?.*$'.format('|'.join(keys)),
        flags=re.IGNORECASE,
    )
    type_tuple = type_re.match(s).groups()
    _type, _args = type_tuple[0].upper(), [int(e) for e in type_tuple[1:] if e]
    return KylinType.get(_type)(*_args)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .kylinpy import Kylin, create_kylin, KylinCluster, dsn_proxy

__version__ = '2.8.4'

__all__ = [
    'Kylin', 'create_kylin', 'KylinCluster', 'dsn_proxy',
]



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.client import InternalServerError, UnauthorizedError
from kylinpy.exceptions import KylinQueryError, KylinJobError
from kylinpy.utils.helper import private_v4_api_warnings
from ._service_interface import ServiceInterface


class _Api(object):
    """
    For testing convenience
    """
    @staticmethod
    def query(client, endpoint, **kwargs):
        return client.post(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def tables_and_columns(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def projects(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def jobs(client, endpoint, params=None, **kwargs):
        return client.get(endpoint=endpoint, params=params, **kwargs).json().get('data')

    @staticmethod
    def resume_job(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def tables(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def model_desc(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def models(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def authentication(client, endpoint, **kwargs):
        rv = client.get(endpoint=endpoint, **kwargs).json().get('data')
        if rv == {}:
            raise UnauthorizedError
        return rv

    @staticmethod
    def build(client, endpoint, **kwargs):
        return client.post(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def build_indexes(client, endpoint, **kwargs):
        return client.post(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def list_index_rules(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def put_index_rules(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def list_indexes(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def delete_index(client, endpoint, **kwargs):
        return client.delete(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def refresh(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def merge(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def list_segment(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json().get('data')

    @staticmethod
    def delete_segment(client, endpoint, **kwargs):
        return client.delete(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def refresh_catalog_cache(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json()


class KE4Service(ServiceInterface):
    api = _Api

    def __init__(self, client, project=None):
        self.client = client
        self.project = project

    def query(self, sql, limit=50000, offset=0, acceptPartial=False, **kwargs):
        json_data = {
            'acceptPartial': acceptPartial,
            'limit': limit,
            'offset': offset,
            'project': self.project,
            'sql': sql,
        }
        kwargs.setdefault('json', json_data)
        try:
            response = self.api.query(self.client, '/query', **kwargs)
        except InternalServerError as err:
            raise KylinQueryError(err)

        err_message = response.get('exceptionMessage')
        if err_message:
            raise KylinQueryError(err_message)

        return response

    def projects(self, **kwargs):
        params = {
            'page_offset': 0,
            'page_size': 1000,
        }
        kwargs.setdefault('params', params)
        _projects = self.api.projects(self.client, '/projects', **kwargs)
        return _projects.get('value')

    def jobs(self, **params):
        _params = {
            'time_filter': 0,
        }
        _params.update(params)
        _jobs = self.api.jobs(self.client, '/jobs', params=_params)
        return _jobs.get('value')

    def job_desc(self, job_id):
        try:
            params = {
                'time_filter': 0,
                'key': str(job_id),
                'project': self.project,
            }
            rv = self.api.jobs(self.client, '/jobs', params=params).get('value')
            if len(rv) > 0:
                return rv[0]
            else:
                raise KylinJobError("Not Find this job {}".format(job_id))
        except InternalServerError as e:
            raise KylinJobError(e)

    def tables_and_columns(self, **kwargs):
        params = {'project': self.project}
        kwargs.setdefault('params', params)
        resp = self.api.tables_and_columns(self.client, '/query/tables_and_columns', **kwargs)
        tbl_pair = tuple(('{}.{}'.format(tbl.get('table_SCHEM'), tbl.get('table_NAME')), tbl) for tbl in resp)
        for tbl in tbl_pair:
            tbl[1]['columns'] = [(col['column_NAME'], col) for col in tbl[1]['columns']]
        return dict(tbl_pair)

    def tables_in_hive(self, **kwargs):
        params = {
            'project': self.project,
            'ext': True,
            'page_offset': 0,
            'page_size': 100000,
        }
        kwargs.setdefault('params', params)
        tables = self.api.tables(self.client, '/tables', **kwargs)
        tables = tables.get('value')

        __tables_in_hive = {}
        for tbl in tables:
            db = tbl['database']
            name = tbl['name']
            fullname = '{}.{}'.format(db, name)
            __tables_in_hive[fullname] = tbl

        return __tables_in_hive

    def model_desc(self, name, **kwargs):
        _url = {
            'project': self.project,
            'model_name': name,
        }
        _model_desc = self.api.model_desc(
            self.client,
            '/models/{project}/{model_name}/model_desc'.format(**_url), **kwargs)
        return _model_desc

    def models(self, **kwargs):
        params = {
            'project': self.project,
            'page_offset': 0,
            'page_size': 1000,
        }
        kwargs.setdefault('params', params)
        _models = self.api.models(self.client, '/models', **kwargs)
        return _models.get('value')

    def get_authentication(self, **kwargs):
        return self.api.authentication(self.client, '/user/authentication', **kwargs)

    def fullbuild(self, model_name):
        json = {
            'project': self.project,
            'start': '1',
            'end': '9223372036854775806',
        }
        endpoint = '/models/{}/segments'.format(model_name)
        return self.api.build(self.client, endpoint, json=json)

    def build(self, model_name, start, end):
        json = {
            'project': self.project,
            'start': str(start),
            'end': str(end),
        }
        endpoint = '/models/{}/segments'.format(model_name)
        return self.api.build(self.client, endpoint, json=json)

    @private_v4_api_warnings
    def get_index_rules_by_model_uuid(self, model_uuid):
        params = {
            'project': self.project,
            'model': model_uuid,
            'page_offset': 0,
            'page_size': 10000,
        }
        return self.api.list_index_rules(self.client, '/index_plans/rule', params=params)

    @private_v4_api_warnings
    def put_index_rules_by_model_uuid(self, model_uuid, load_data=False, aggregation_groups=None):
        _json = {
            'aggregation_groups': aggregation_groups if aggregation_groups else [],
            'load_data': load_data,
            'model_id': model_uuid,
            'project': self.project,
        }
        return self.api.put_index_rules(self.client, '/index_plans/rule', json=_json)

    @private_v4_api_warnings
    def get_indexes_by_model_uuid(self, model_uuid):
        params = {
            'project': self.project,
            'model': model_uuid,
            'page_offset': 0,
            'page_size': 10000,
        }
        rv = self.api.list_indexes(self.client, '/index_plans/index', params=params)
        return rv.get('value', [])

    def build_indexes(self, model_name):
        json = {
            'project': self.project,
        }
        endpoint = '/models/{}/indexes'.format(model_name)
        return self.api.build_indexes(self.client, endpoint, json=json)

    @private_v4_api_warnings
    def delete_index(self, model_uuid, index_id):
        params = {
            'project': self.project,
            'model': model_uuid,
        }
        return self.api.delete_index(self.client, '/index_plans/index/{}'.format(index_id), params=params)

    def refresh(self, model_name, ids):
        json = {
            'project': self.project,
            'type': 'REFRESH',
            'ids': ids,
        }
        res = self.api.refresh(self.client, '/models/{}/segments'.format(model_name), json=json)
        return res

    def merge(self, model_name, ids):
        json = {
            'project': self.project,
            'type': 'MERGE',
            'ids': ids,
        }
        res = self.api.merge(self.client, '/models/{}/segments'.format(model_name), json=json)
        return res

    def list_segment(self, model_name):
        params = {
            'project': self.project,
            'model_name': model_name,
            'page_size': 2147483646,
        }
        endpoint = '/models/{}/segments'.format(model_name)
        return self.api.list_segment(self.client, endpoint, params=params)

    def delete_segment(self, model_name, ids, **kwargs):
        params = {
            'project': self.project,
            'ids': ids,
        }
        params.update(kwargs)
        endpoint = '/models/{}/segments'.format(model_name)
        return self.api.delete_segment(self.client, endpoint, params=params)

    @private_v4_api_warnings
    def refresh_catalog_cache(self, tables):
        params = {
            'tables': tables,
        }
        endpoint = '/tables/catalog_cache'
        return self.api.refresh_catalog_cache(self.client, endpoint, params=params)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from datetime import datetime

import pytest

from kylinpy import create_kylin
from kylinpy.exceptions import KylinCubeError


class TestCubeSource:
    def test_kylin_cube_source(self, v1_api):
        project = create_kylin('kylin://username:password@example/kylin_sales')
        assert hasattr(project, 'get_cube_source')
        cube = project.get_datasource('kylin_sales_cube')
        assert cube.name == 'kylin_sales_cube'
        assert cube.model_name == 'kylin_sales_model'
        assert cube.cube_name == 'kylin_sales_cube'

        # fact table testing
        assert cube.fact_table.alias == 'KYLIN_SALES'
        assert cube.fact_table.fullname == 'DEFAULT.KYLIN_SALES'
        assert cube.fact_table.scheme == 'DEFAULT'
        assert cube.fact_table.name == 'KYLIN_SALES'

        assert cube.identity == '0ef9b7a8-3929-4dff-b59d-2100aadc8dbf'
        assert cube.last_modified == 1560061109000

        assert [d.name for d in cube.dimensions] == [
            'KYLIN_SALES.TRANS_ID',
            'KYLIN_CAL_DT.YEAR_BEG_DT',
            'KYLIN_CAL_DT.MONTH_BEG_DT',
            'KYLIN_CAL_DT.WEEK_BEG_DT',
            'KYLIN_CATEGORY_GROUPINGS.USER_DEFINED_FIELD1',
            'KYLIN_CATEGORY_GROUPINGS.USER_DEFINED_FIELD3',
            'KYLIN_CATEGORY_GROUPINGS.META_CATEG_NAME',
            'KYLIN_CATEGORY_GROUPINGS.CATEG_LVL2_NAME',
            'KYLIN_CATEGORY_GROUPINGS.CATEG_LVL3_NAME',
            'KYLIN_SALES.LSTG_FORMAT_NAME',
            'KYLIN_SALES.SELLER_ID',
            'KYLIN_SALES.BUYER_ID',
            'BUYER_ACCOUNT.ACCOUNT_BUYER_LEVEL',
            'SELLER_ACCOUNT.ACCOUNT_SELLER_LEVEL',
            'BUYER_ACCOUNT.ACCOUNT_COUNTRY',
            'SELLER_ACCOUNT.ACCOUNT_COUNTRY',
            'BUYER_COUNTRY.NAME',
            'SELLER_COUNTRY.NAME',
            'KYLIN_SALES.OPS_USER_ID',
            'KYLIN_SALES.OPS_REGION',
        ]

        assert [m.name for m in cube.measures] == [
            'GMV_SUM',
            'BUYER_LEVEL_SUM',
            'SELLER_LEVEL_SUM',
            'TRANS_CNT',
            'SELLER_CNT_HLL',
            'TOP_SELLER',
        ]

        assert str(cube.from_clause) == (
            '"DEFAULT"."KYLIN_SALES" AS "KYLIN_SALES" '
            'JOIN "DEFAULT"."KYLIN_CAL_DT" AS "KYLIN_CAL_DT" ON "KYLIN_SALES"."PART_DT" = "KYLIN_CAL_DT"."CAL_DT" '  # noqa
            'JOIN "DEFAULT"."KYLIN_CATEGORY_GROUPINGS" AS "KYLIN_CATEGORY_GROUPINGS" ON "KYLIN_SALES"."LEAF_CATEG_ID" = "KYLIN_CATEGORY_GROUPINGS"."LEAF_CATEG_ID" AND "KYLIN_SALES"."LSTG_SITE_ID" = "KYLIN_CATEGORY_GROUPINGS"."SITE_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_ACCOUNT" AS "BUYER_ACCOUNT" ON "KYLIN_SALES"."BUYER_ID" = "BUYER_ACCOUNT"."ACCOUNT_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_ACCOUNT" AS "SELLER_ACCOUNT" ON "KYLIN_SALES"."SELLER_ID" = "SELLER_ACCOUNT"."ACCOUNT_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_COUNTRY" AS "BUYER_COUNTRY" ON "BUYER_ACCOUNT"."ACCOUNT_COUNTRY" = "BUYER_COUNTRY"."COUNTRY" '  # noqa
            'JOIN "DEFAULT"."KYLIN_COUNTRY" AS "SELLER_COUNTRY" ON "SELLER_ACCOUNT"."ACCOUNT_COUNTRY" = "SELLER_COUNTRY"."COUNTRY"'  # noqa
        )
        assert cube.support_invoke_command == {
            'fullbuild', 'build', 'merge', 'refresh',
            'delete', 'build_streaming', 'merge_streaming', 'refresh_streaming',
            'disable', 'enable', 'purge', 'clone', 'drop',
        }
        assert cube.fullbuild() == {'build': 'success'}
        assert cube.build(datetime(2000, 1, 1), datetime(2000, 1, 2)) == {'build': 'success'}
        assert cube.list_segment()[0].get('uuid') == 'b5999bec-2381-77c7-cafb-c59407a7a032'
        assert cube.merge(datetime(2000, 1, 1), datetime(2000, 1, 2)) == {'build': 'success'}
        assert cube.refresh(datetime(2000, 1, 1), datetime(2000, 1, 2)) == {'build': 'success'}
        assert cube.delete('segment_name') == {'success': 'success'}
        assert cube.build_streaming(123, 456) == {'success': 'success'}
        assert cube.merge_streaming(123, 456) == {'success': 'success'}
        assert cube.refresh_streaming(123, 456) == {'success': 'success'}
        assert cube.disable() == {'success': 'success'}
        assert cube.enable() == {'success': 'success'}
        assert cube.purge() == {'success': 'success'}
        assert cube.clone() == {'success': 'success'}
        assert cube.drop() == {'success': 'success'}
        with pytest.raises(KylinCubeError):
            cube.invoke_command('invalid command')
        build_cmd = cube.invoke_command('build', foo='unused', bar='unused',
                                        start=datetime(2000, 1, 1), end=datetime(2000, 1, 1))
        drop_cmd = cube.invoke_command('drop')
        assert build_cmd == {'build': 'success'}
        assert drop_cmd == {'success': 'success'}

    def test_v2_cube_source(self, v2_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v2')
        cube = project.get_cube_source('kylin_sales_cube')
        assert cube.name == 'kylin_sales_cube'
        assert cube.model_name == 'kylin_sales_model'
        assert cube.cube_name == 'kylin_sales_cube'

        # fact table testing
        assert cube.fact_table.alias == 'KYLIN_SALES'
        assert cube.fact_table.fullname == 'DEFAULT.KYLIN_SALES'
        assert cube.fact_table.scheme == 'DEFAULT'
        assert cube.fact_table.name == 'KYLIN_SALES'

        assert cube.identity == '0ef9b7a8-3929-4dff-b59d-2100aadc8dbf'
        assert cube.last_modified == 1573799855000

        assert [d.name for d in cube.dimensions] == [
            'KYLIN_SALES.TRANS_ID',
            'KYLIN_CAL_DT.YEAR_BEG_DT',
            'KYLIN_CAL_DT.MONTH_BEG_DT',
            'KYLIN_CAL_DT.WEEK_BEG_DT',
            'KYLIN_CATEGORY_GROUPINGS.USER_DEFINED_FIELD1',
            'KYLIN_CATEGORY_GROUPINGS.USER_DEFINED_FIELD3',
            'KYLIN_CATEGORY_GROUPINGS.META_CATEG_NAME',
            'KYLIN_CATEGORY_GROUPINGS.CATEG_LVL2_NAME',
            'KYLIN_CATEGORY_GROUPINGS.CATEG_LVL3_NAME',
            'KYLIN_SALES.LSTG_FORMAT_NAME',
            'KYLIN_SALES.SELLER_ID',
            'KYLIN_SALES.BUYER_ID',
            'BUYER_ACCOUNT.ACCOUNT_BUYER_LEVEL',
            'SELLER_ACCOUNT.ACCOUNT_SELLER_LEVEL',
            'BUYER_ACCOUNT.ACCOUNT_COUNTRY',
            'SELLER_ACCOUNT.ACCOUNT_COUNTRY',
            'BUYER_COUNTRY.NAME',
            'SELLER_COUNTRY.NAME',
            'KYLIN_SALES.OPS_USER_ID',
            'KYLIN_SALES.OPS_REGION',
        ]
        assert [d.column.name for d in cube.dimensions] == [
            'TRANS_ID',
            'YEAR_BEG_DT',
            'MONTH_BEG_DT',
            'WEEK_BEG_DT',
            'USER_DEFINED_FIELD1',
            'USER_DEFINED_FIELD3',
            'META_CATEG_NAME',
            'CATEG_LVL2_NAME',
            'CATEG_LVL3_NAME',
            'LSTG_FORMAT_NAME',
            'SELLER_ID',
            'BUYER_ID',
            'ACCOUNT_BUYER_LEVEL',
            'ACCOUNT_SELLER_LEVEL',
            'ACCOUNT_COUNTRY',
            'ACCOUNT_COUNTRY',
            'NAME',
            'NAME',
            'OPS_USER_ID',
            'OPS_REGION',
        ]
        assert [d.column.alias for d in cube.dimensions] == [
            'TRANS_ID',
            'YEAR_BEG_DT',
            'MONTH_BEG_DT',
            'WEEK_BEG_DT',
            'USER_DEFINED_FIELD1',
            'USER_DEFINED_FIELD3',
            'META_CATEG_NAME',
            'CATEG_LVL2_NAME',
            'CATEG_LVL3_NAME',
            'LSTG_FORMAT_NAME',
            'SELLER_ID',
            'BUYER_ID',
            'ACCOUNT_BUYER_LEVEL',
            'ACCOUNT_SELLER_LEVEL',
            'BUYER_COUNTRY',
            'SELLER_COUNTRY',
            'BUYER_COUNTRY_NAME',
            'SELLER_COUNTRY_NAME',
            'OPS_USER_ID',
            'OPS_REGION',
        ]
        assert [d.column.datatype for d in cube.dimensions] == [
            'BIGINT',
            'DATE',
            'DATE',
            'DATE',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'BIGINT',
            'BIGINT',
            'INTEGER',
            'INTEGER',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
        ]

        assert [m.name for m in cube.measures] == [
            'GMV_SUM',
            'BUYER_LEVEL_SUM',
            'SELLER_LEVEL_SUM',
            'TRANS_CNT',
            'SELLER_CNT_HLL',
            'TOP_SELLER',
        ]

        assert str(cube.from_clause) == (
            '"DEFAULT"."KYLIN_SALES" AS "KYLIN_SALES" '
            'JOIN "DEFAULT"."KYLIN_CAL_DT" AS "KYLIN_CAL_DT" ON "KYLIN_SALES"."PART_DT" = "KYLIN_CAL_DT"."CAL_DT" '  # noqa
            'JOIN "DEFAULT"."KYLIN_CATEGORY_GROUPINGS" AS "KYLIN_CATEGORY_GROUPINGS" ON "KYLIN_SALES"."LEAF_CATEG_ID" = "KYLIN_CATEGORY_GROUPINGS"."LEAF_CATEG_ID" AND "KYLIN_SALES"."LSTG_SITE_ID" = "KYLIN_CATEGORY_GROUPINGS"."SITE_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_ACCOUNT" AS "BUYER_ACCOUNT" ON "KYLIN_SALES"."BUYER_ID" = "BUYER_ACCOUNT"."ACCOUNT_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_ACCOUNT" AS "SELLER_ACCOUNT" ON "KYLIN_SALES"."SELLER_ID" = "SELLER_ACCOUNT"."ACCOUNT_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_COUNTRY" AS "BUYER_COUNTRY" ON "BUYER_ACCOUNT"."ACCOUNT_COUNTRY" = "BUYER_COUNTRY"."COUNTRY" '  # noqa
            'JOIN "DEFAULT"."KYLIN_COUNTRY" AS "SELLER_COUNTRY" ON "SELLER_ACCOUNT"."ACCOUNT_COUNTRY" = "SELLER_COUNTRY"."COUNTRY"'  # noqa
        )



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.exceptions import UnsupportApiError
from ._job_interface import JobInterface


class Ke4Job(JobInterface):
    def __init__(self, job_id, service):
        self.job_id = job_id
        self.service = service

    def resume(self):
        raise UnsupportApiError("Now ke4 jobs does not support resume")

    def cancel(self):
        raise UnsupportApiError("Now ke4 jobs does not support cancel")

    def get_desc(self):
        return self.service.job_desc(self.job_id)

    def pause(self):
        raise UnsupportApiError("Now ke4 jobs does not support pause")

    def drop(self):
        raise UnsupportApiError("Now ke4 jobs does not support drop")

    @property
    def status(self):
        desc = self.get_desc()
        if desc:
            return desc.get('job_status')
        return None

    def __repr__(self):
        return '<KylinJob instance: {}>'.format(self.job_id)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from kylinpy.client import InternalServerError, UnauthorizedError
from kylinpy.exceptions import KylinQueryError, KylinCubeError, KylinJobError
from ._service_interface import ServiceInterface


class _Api(object):
    """
    For testing convenience
    """
    @staticmethod
    def query(client, endpoint, **kwargs):
        return client.post(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def tables_and_columns(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def projects(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def jobs(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def maintain_job(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def drop_job(client, endpoint, **kwargs):
        return client.delete(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def job_desc(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def tables(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def cube_desc(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def models(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def cubes(client, endpoint, **kwargs):
        return client.get(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def authentication(client, endpoint, **kwargs):
        rv = client.get(endpoint=endpoint, **kwargs).json()
        if rv == {}:
            raise UnauthorizedError
        return rv

    @staticmethod
    def build(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def build_streaming(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def delete_segment(client, endpoint, **kwargs):
        return client.delete(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def maintain_cube(client, endpoint, **kwargs):
        return client.put(endpoint=endpoint, **kwargs).json()

    @staticmethod
    def drop_cube(client, endpoint, **kwargs):
        return client.delete(endpoint=endpoint, **kwargs).json()


class KylinService(ServiceInterface):
    api = _Api

    def __init__(self, client, project=None):
        self.client = client
        self.project = project

    def query(self, sql, limit=50000, offset=0, acceptPartial=False, **kwargs):
        json_data = {
            'acceptPartial': acceptPartial,
            'limit': limit,
            'offset': offset,
            'project': self.project,
            'sql': sql,
        }
        kwargs.setdefault('json', json_data)
        try:
            response = self.api.query(self.client, '/query', **kwargs)
        except InternalServerError as err:
            raise KylinQueryError(err)

        err_message = response.get('exceptionMessage')
        if err_message:
            raise KylinQueryError(err_message)

        return response

    def projects(self, **kwargs):
        params = {
            'pageOffset': 0,
            'pageSize': 1000,
        }
        kwargs.setdefault('params', params)
        _projects = self.api.projects(self.client, '/projects', **kwargs)
        return _projects

    def jobs(self, **params):
        _params = {
            'projectName': self.project,
            'limit': 15,
            'offset': 0,
        }
        _params.update(params)
        return self.api.jobs(self.client, '/jobs', params=_params)

    def maintain_job(self, job_id, maintain_type):
        if maintain_type not in ('resume', 'cancel', 'pause'):
            raise KylinJobError(
                "Unsupported maintain type: {}, "
                "The maintain type must be 'resume', 'cancel',  'pause'".format(maintain_type))

        try:
            return self.api.maintain_job(self.client, '/jobs/{}/{}'.format(job_id, maintain_type))
        except InternalServerError as e:
            raise KylinJobError(e)

    def drop_job(self, job_id):
        try:
            return self.api.drop_job(self.client, '/jobs/{}/drop'.format(job_id))
        except InternalServerError as e:
            raise KylinJobError(e)

    def job_desc(self, job_id):
        try:
            return self.api.job_desc(self.client, '/jobs/{}'.format(job_id))
        except InternalServerError as e:
            raise KylinJobError(e)

    def tables_and_columns(self, **kwargs):
        params = {
            'project': self.project,
        }
        kwargs.setdefault('params', params)
        resp = self.api.tables_and_columns(self.client, '/tables_and_columns', **kwargs)
        tbl_pair = tuple(('{}.{}'.format(tbl.get('table_SCHEM'), tbl.get('table_NAME')), tbl) for tbl in resp)
        for tbl in tbl_pair:
            tbl[1]['columns'] = [(col['column_NAME'], col) for col in tbl[1]['columns']]
        return dict(tbl_pair)

    def tables_in_hive(self, **kwargs):
        params = {
            'project': self.project,
            'ext': True,
        }
        kwargs.setdefault('params', params)
        tables = self.api.tables(self.client, '/tables', **kwargs)

        __tables_in_hive = {}
        for tbl in tables:
            db = tbl['database']
            name = tbl['name']
            fullname = '{}.{}'.format(db, name)
            __tables_in_hive[fullname] = tbl

        return __tables_in_hive

    def cube_desc(self, name, **kwargs):
        return self.api.cube_desc(self.client, '/cube_desc/{}/desc'.format(name), **kwargs)

    def model_desc(self, name, **kwargs):
        return [_ for _ in self.models(**kwargs) if _.get('name') == name][0]

    def models(self, **kwargs):
        params = {
            'projectName': self.project,
            'pageOffset': 0,
            'pageSize': 1000,
        }
        kwargs.setdefault('params', params)
        return self.api.models(self.client, '/models', **kwargs)

    def cubes(self, name=None, **kwargs):
        params = {
            'pageOffset': 0,
            'offset': 0,
            'limit': 50000,
            'pageSize': 1000,
            'projectName': self.project,
        }
        if name:
            params.update({'cubeName': name})
        kwargs.setdefault('params', params)
        return self.api.cubes(self.client, '/cubes', **kwargs)

    def get_authentication(self, **kwargs):
        rv = self.api.authentication(self.client, '/user/authentication', **kwargs)
        return rv.get('userDetails')

    def fullbuild(self, cube_name):
        json = {
            'startTime': 0,
            'endTime': 0,
            'buildType': 'BUILD',
        }
        endpoint = '/cubes/{}/build'.format(cube_name)
        return self.api.build(self.client, endpoint, json=json)

    def build(self, cube_name, build_type, start, end):
        if build_type not in ('BUILD', 'MERGE', 'REFRESH'):
            raise KylinCubeError(
                "Unsupported build type: {}, The build type must be 'BUILD', 'MERGE', 'REFRESH'".format(build_type))

        json = {
            'startTime': start,
            'endTime': end,
            'buildType': build_type,
        }
        endpoint = '/cubes/{}/build'.format(cube_name)
        return self.api.build(self.client, endpoint, json=json)

    def build_streaming(self, cube_name, build_type, offset_start, offset_end):
        if build_type not in ('BUILD', 'MERGE', 'REFRESH'):
            raise KylinCubeError(
                "Unsupported build type: {}, The build type must be 'BUILD', 'MERGE', 'REFRESH'".format(build_type))

        json = {
            'sourceOffsetStart': offset_start,
            'sourceOffsetEnd': offset_end,
            'buildType': build_type,
        }
        endpoint = '/cubes/{}/build2'.format(cube_name)
        return self.api.build_streaming(self.client, endpoint, json=json)

    def delete_segment(self, cube_name, segment_name):
        endpoint = '/cubes/{}/segs/{}'.format(cube_name, segment_name)
        return self.api.delete_segment(self.client, endpoint)

    def maintain_cube(self, cube_name, maintain_type):
        if maintain_type not in ('disable', 'enable', 'purge', 'clone'):
            raise KylinCubeError(
                "Unsupported maintain type: {}, "
                "The maintain type must be 'BUILD', 'MERGE', 'REFRESH'".format(maintain_type))

        endpoint = '/cubes/{}/{}'.format(cube_name, maintain_type)
        json = {}
        if maintain_type == 'clone':
            json = {
                'cubeName': '{}_clone'.format(cube_name),
                'project': self.project,
            }
        try:
            return self.api.maintain_cube(self.client, endpoint, json=json)
        except InternalServerError as e:
            raise KylinCubeError(e)

    def drop_cube(self, cube_name):
        endpoint = '/cubes/{}'.format(cube_name)
        try:
            return self.api.drop_cube(self.client, endpoint)
        except InternalServerError as e:
            raise KylinCubeError(e)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .kylin_job_object import KylinJob  # noqa
from .ke3_job_object import Ke3Job  # noqa
from .ke4_job_object import Ke4Job  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class JobInterface(object):

    def pause(self):
        raise NotImplementedError()

    def resume(self):
        raise NotImplementedError()

    def cancel(self):
        raise NotImplementedError()

    def get_desc(self):
        raise NotImplementedError()

    def drop(self):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json

import pytest

from kylinpy.client import Client
from kylinpy.client.exceptions import HTTPError
from kylinpy.utils.compat import urllib


class MockException(HTTPError):
    def __init__(self, code):
        self.code = code
        self.reason = 'REASON'
        self.hdrs = 'HEADERS'

    def read(self):
        return 'BODY'


class MockResponse(urllib.HTTPSHandler):
    def __init__(self, response_code):
        self.response_code = response_code
        self.url = '/foo/bar'

    def getcode(self):
        return self.response_code

    def info(self):
        return 'HEADERS'

    def read(self):
        return json.dumps({'hello': 'world'}).encode('utf-8')


class TestClient(object):
    @pytest.fixture
    def client(self):
        host = 'http://example/'
        request_headers = {'Content-Type': 'application/json'}
        client = Client(host=host, request_headers=request_headers)
        yield client

    def test_init(self, client):
        assert client.host == 'http://example'
        assert client.request_headers == {'Content-Type': 'application/json'}
        assert client.prefix is None
        assert client.timeout is None
        assert client.unverified is None
        assert client.is_debug is False

    def test_get(self, client, mocker):
        mocker.patch('kylinpy.client.client.Client._make_request', return_value=MockResponse(200))

        rv = client.get('/foo/bar', params={'param1': 'x', 'param2': 'y'})
        assert rv.url == '/foo/bar'
        assert rv.status_code == 200
        assert rv.headers == 'HEADERS'
        assert rv.json() == {'hello': 'world'}

    def test_post(self, client, mocker):
        mocker.patch('kylinpy.client.client.Client._make_request', return_value=MockResponse(200))

        rv = client.post('/foo/bar', params={'param1': 'x', 'param2': 'y'})
        assert rv.url == '/foo/bar'
        assert rv.status_code == 200
        assert rv.headers == 'HEADERS'
        assert rv.json() == {'hello': 'world'}

    def test_put(self, client, mocker):
        mocker.patch('kylinpy.client.client.Client._make_request', return_value=MockResponse(200))

        rv = client.post('/foo/bar', params={'param1': 'x', 'param2': 'y'})
        assert rv.url == '/foo/bar'
        assert rv.status_code == 200
        assert rv.headers == 'HEADERS'
        assert rv.json() == {'hello': 'world'}

    def test_delete(self, client, mocker):
        mocker.patch('kylinpy.client.client.Client._make_request', return_value=MockResponse(200))

        rv = client.delete('/foo/bar', params={'param1': 'x', 'param2': 'y'})
        assert rv.url == '/foo/bar'
        assert rv.status_code == 200
        assert rv.headers == 'HEADERS'
        assert rv.json() == {'hello': 'world'}



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import pytest

from kylinpy.job import KylinJob
from kylinpy.kylinpy import create_kylin, SERVICES
from kylinpy.client import Client
from kylinpy.exceptions import NoSuchTableError


class TestProject(object):
    @property
    def project(self):
        return create_kylin('kylin://username:password@example/foobar')

    def test_init(self):
        cluster = create_kylin('kylin://name@45中文:pwd12@%+@example.com:9000/foobar')
        assert cluster.host == 'example.com'
        assert cluster.port == 9000
        assert cluster.username == 'name@45中文'
        assert cluster.password == 'pwd12@%+'
        assert cluster.is_ssl is False
        assert cluster.prefix == '/kylin/api'
        assert cluster.timeout == 30
        assert cluster.unverified is True
        assert cluster.version == 'v1'
        assert cluster.is_pushdown is False
        assert cluster.is_pushdown is False
        assert cluster.is_debug is False
        assert cluster.scheme == 'http'
        assert cluster.project == 'foobar'

    def test_service(self):
        cluster = create_kylin('kylin://foo:bar@example.com:9000/foobar?prefix=/aaa/bbb')
        assert isinstance(cluster.service, SERVICES['v1'])
        assert cluster.service.client.request_headers.get('Accept') is None
        assert cluster.service.client.request_headers.get('Authorization') == 'Basic Zm9vOmJhcg=='
        assert cluster.service.project == 'foobar'
        assert cluster.version == 'v1'
        cluster.username = 'yongjie'
        cluster.password = 'zhao'
        cluster.host = 'eager_host.com'
        cluster.port = 1234
        cluster.scheme = 'https'
        cluster.timeout = 1000
        cluster.unverified = False
        cluster.prefix = '/hello/world'
        cluster.is_debug = True
        eager_client = cluster.service.client
        assert eager_client.host == 'https://eager_host.com:1234'
        assert eager_client.prefix == '/hello/world'
        assert eager_client.unverified is False
        assert eager_client.timeout == 1000
        assert eager_client.is_debug is True
        assert cluster.service.client.request_headers.get('Authorization') == 'Basic eW9uZ2ppZTp6aGFv'

        cluster2 = create_kylin('kylin://foo:bar@example.com:9000/?version=v2')
        assert cluster2.version == 'v2'
        assert cluster2.service.client.request_headers.get('Accept') == 'application/vnd.apache.kylin-v2+json'
        assert isinstance(cluster2.service, SERVICES['v2'])

        cluster4 = create_kylin('kylin://foo:bar@example.com:9000/?version=v4')
        assert cluster4.version == 'v4'
        assert cluster4.service.client.request_headers.get('Accept') == 'application/vnd.apache.kylin-v4-public+json'
        assert isinstance(cluster4.service, SERVICES['v4'])

    def test_get_client(self):
        cluster = create_kylin('kylin://username:password@example')
        assert isinstance(cluster._get_client(), Client)

    def test_basic_auth_dump(self):
        cluster = create_kylin('kylin://username:password@example')
        assert cluster.basic_auth_dump('foo', 'bar') == {'Authorization': 'Basic Zm9vOmJhcg=='}

    def test_query(self, v1_api):
        rv = self.project.query('select count(*) from kylin_sales')
        assert 'columnMetas' in rv
        assert rv['results'] == [['10000']]

    def test_get_all_tables(self, v1_api):
        assert self.project.get_all_tables() == [
            'KYLIN_ACCOUNT',
            'KYLIN_CAL_DT',
            'KYLIN_CATEGORY_GROUPINGS',
            'KYLIN_COUNTRY',
            'KYLIN_SALES',
        ]

        pushdown = create_kylin('kylin://username:password@example/foobar?is_pushdown=1')
        assert pushdown.is_pushdown is True
        assert pushdown.get_all_tables() == [
            'KYLIN_ACCOUNT',
            'KYLIN_CAL_DT',
            'KYLIN_CATEGORY_GROUPINGS',
            'KYLIN_COUNTRY',
            'KYLIN_SALES',
            'KYLIN_STREAMING_TABLE',
        ]

    def test_get_all_schema(self, v1_api):
        assert self.project.get_all_schemas() == ['DEFAULT']

    def test_get_table_source(self, v1_api):
        table = self.project.get_table_source('KYLIN_SALES', 'DEFAULT')
        assert table.name == 'KYLIN_SALES'
        assert table.schema == 'DEFAULT'

    def test_get_table_source_with_schema(self, v1_api):
        table = self.project.get_table_source('DEFAULT.KYLIN_SALES')
        assert table.name == 'KYLIN_SALES'
        assert table.schema == 'DEFAULT'

    def test_get_table_source_error(self, v1_api):
        with pytest.raises(NoSuchTableError):
            self.project.get_table_source('foo.bar.zee.hello.world')

    def test_get_cube_source(self, v1_api):
        cube = self.project.get_cube_source('kylin_sales_cube')
        assert cube.name == 'kylin_sales_cube'
        assert cube.model_name == 'kylin_sales_model'

    def test_get_datasource(self, v1_api):
        cube = self.project.get_cube_source('kylin_sales_cube')
        assert cube.name == 'kylin_sales_cube'
        assert cube.model_name == 'kylin_sales_model'

    def test_list_job(self, v1_api):
        job_list = self.project.list_job()
        assert isinstance(job_list[0], KylinJob)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import base64
import logging
import warnings

from kylinpy.client import Client as HTTPClient
from kylinpy.exceptions import KylinCubeError
from kylinpy.service import KylinService, KE3Service, KE4Service
from kylinpy.datasource import TableSource, CubeSource, KE4ModelSource
from kylinpy.job import KylinJob, Ke3Job, Ke4Job
from kylinpy.utils.compat import as_unicode, urlparse, parse_qsl

SERVICES = {
    'v1': KylinService,
    'v2': KE3Service,
    'v4': KE4Service,
}

Job = {
    'v1': KylinJob,
    'v2': Ke3Job,
    'v4': Ke4Job,
}


class Kylin(object):
    def __init__(self, host, username=None, password=None, port=7070, project=None, **connect_args):
        if host.startswith(('http://', 'https://')):
            _, self.host = host.split('://')
        else:
            self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.is_ssl = connect_args.get('is_ssl', False)
        self.prefix = connect_args.get('prefix', '/kylin/api')
        self.timeout = int(connect_args.get('timeout', 30))
        self.unverified = bool(connect_args.get('unverified', True))
        self.version = connect_args.get('version', 'v1')
        self.is_pushdown = bool(connect_args.get('is_pushdown', False))
        self.is_debug = bool(connect_args.get('is_debug', False))
        self.scheme = 'https' if self.is_ssl else 'http'
        self.project = project
        if self.is_debug:
            logging.basicConfig(level=logging.DEBUG)

    @property
    def service(self):
        return SERVICES[self.version](self._get_client(), self.project)

    def _get_client(self):
        headers = {
            'User-Agent': 'Kylin Python Client',
        }

        if self.version == 'v2':
            headers.update({'Accept': 'application/vnd.apache.kylin-v2+json'})

        if self.version == 'v4':
            headers.update({'Accept': 'application/vnd.apache.kylin-v4-public+json'})

        if self.username and self.password:
            headers.update(self.basic_auth_dump(self.username, self.password))

        return HTTPClient(
            host='{self.scheme}://{self.host}:{self.port}'.format(**locals()),
            prefix=self.prefix,
            timeout=self.timeout,
            request_headers=headers,
            unverified=self.unverified,
            is_debug=self.is_debug,
        )

    def basic_auth_dump(self, username, password):
        _auth = as_unicode('{}:{}').format(username, password)
        _auth = base64.b64encode(_auth.encode('utf-8')).decode('ascii')
        return {'Authorization': 'Basic {}'.format(_auth)}

    def projects(self):
        return self.service.projects()

    def query(self, sql, **parameters):
        return self.service.query(sql, **parameters)

    def get_all_tables(self, schema=None):
        if self.is_pushdown:
            _full_names = sorted(list(self.service.tables_in_hive().keys()))
        else:
            _full_names = sorted(list(self.service.tables_and_columns().keys()))

        if schema:
            _full_names = [t for t in _full_names if t.split('.')[0] == schema]
        return [t.split('.')[1] for t in _full_names]

    def get_all_schemas(self):
        if self.is_pushdown:
            _full_names = sorted(list(self.service.tables_in_hive().keys()))
        else:
            _full_names = sorted(list(self.service.tables_and_columns().keys()))
        return list(set(t.split('.')[0] for t in _full_names))

    def get_table_source(self, name, schema=None):
        if '.' in name:
            schema, name = name.split('.', 1)
        fullname = '{}.{}'.format(schema, name)
        if self.is_pushdown:
            return TableSource(name, schema, self.service.tables_in_hive().get(fullname))
        else:
            return TableSource(name, schema, self.service.tables_and_columns().get(fullname))

    def get_cube_source(self, name):
        warnings.warn('This method is deprecated. Please use `get_datsource()`.')
        return self.get_datasource(name)

    def get_datasource(self, name):
        if self.version == 'v4':
            _params = {
                'project': self.project,
                'page_offset': 0,
                'page_size': 1000,
                'model_name': name,
                'exact': True,
            }
            model_desc = self.service.models(params=_params)[0]
            return KE4ModelSource(
                model_desc=model_desc,
                tables_and_columns=self.service.tables_and_columns(),
                service=self.service,
            )

        cube_desc = self.service.cube_desc(name)
        if cube_desc is None:
            raise KylinCubeError('No Cube found: {}'.format(name))
        model_name = cube_desc.get('model_name')
        return CubeSource(
            cube_desc=cube_desc,
            model_desc=self.service.model_desc(model_name),
            tables_and_columns=self.service.tables_and_columns(),
            service=self.service,
        )

    def list_datasources(self):
        if self.version == 'v4':
            return [models.get('name') for models in self.service.models()]
        return [cube.get('name') for cube in self.service.cubes()]

    def get_job(self, job_id):
        return Job[self.version](job_id=job_id, service=self.service)

    def list_job(self, **query_params):
        jobs = self.service.jobs(params=query_params)
        if self.version == 'v4':
            return [self.get_job(job_id=job['id']) for job in jobs]
        else:
            return [self.get_job(job_id=job['uuid']) for job in jobs]

    def __str__(self):
        if self.project:
            dsn = ('{self.scheme}://'
                   '{self.username}:{self.password}@'
                   '{self.host}:{self.port}'
                   '/{self.project}')
        else:
            dsn = ('{self.scheme}://'
                   '{self.username}:{self.password}@'
                   '{self.host}:{self.port}')
        return dsn.format(**locals())

    def __repr__(self):
        return '<Kylinpy instance: {}>'.format(str(self))


def create_kylin(dsn):
    url = urlparse(dsn)
    project = url.path.lstrip('/')
    port = url.port or 7070
    query = dict(parse_qsl(url.query) or {})
    return Kylin(url.hostname, url.username, url.password, port, project, **query)


# For compatibility with older versions
KylinCluster = Kylin
dsn_proxy = create_kylin



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from datetime import datetime

from kylinpy import create_kylin


class TestKe4ModelSource:
    def test_v4_model_source(self, v4_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v4')
        model = project.get_datasource('kylin_sales_model')
        assert model.name == 'kylin_sales_model'
        assert model.model_name == 'kylin_sales_model'

        # fact table testing
        assert model.fact_table.alias == 'KYLIN_SALES'
        assert model.fact_table.fullname == 'DEFAULT.KYLIN_SALES'
        assert model.fact_table.scheme == 'DEFAULT'
        assert model.fact_table.name == 'KYLIN_SALES'

        assert model.identity == '0e788bb6-d56c-44fd-8fe0-26bb77aa40c5'
        assert model.uuid == '0e788bb6-d56c-44fd-8fe0-26bb77aa40c5'
        assert model.last_modified == 1581343542414
        assert model.support_invoke_command == {
            'fullbuild', 'build', 'merge', 'refresh', 'delete', 'list_segment', 'refresh_catalog_cache',
        }

        assert [d.name for d in model.dimensions] == [
            'KYLIN_SALES.PART_DT',
            'KYLIN_SALES.LSTG_FORMAT_NAME',
            'KYLIN_SALES.OPS_REGION',
            'KYLIN_CAL_DT.MONTH_BEG_DT',
            'KYLIN_CAL_DT.CAL_DT',
            'KYLIN_CAL_DT.YEAR_BEG_DT',
            'KYLIN_CAL_DT.QTR_BEG_DT',
            'BUYER_ACCOUNT.ACCOUNT_COUNTRY',
            'BUYER_ACCOUNT.ACCOUNT_BUYER_LEVEL',
            'BUYER_COUNTRY.NAME',
            'BUYER_COUNTRY.COUNTRY',
            'SELLER_ACCOUNT.ACCOUNT_SELLER_LEVEL',
            'SELLER_ACCOUNT.ACCOUNT_COUNTRY',
            'SELLER_COUNTRY.NAME',
            'SELLER_COUNTRY.COUNTRY',
            'KYLIN_CATEGORY_GROUPINGS.CATEG_LVL3_NAME',
            'KYLIN_CATEGORY_GROUPINGS.CATEG_LVL2_NAME',
            'KYLIN_CATEGORY_GROUPINGS.LEAF_CATEG_NAME',
            'KYLIN_CATEGORY_GROUPINGS.META_CATEG_NAME',
        ]
        assert [d.verbose for d in model.dimensions] == [
            'KYLIN_SALES_PART_DT',
            'KYLIN_SALES_LSTG_FORMAT_NAME',
            'KYLIN_SALES_OPS_REGION',
            'KYLIN_CAL_DT_MONTH_BEG_DT',
            'KYLIN_CAL_DT_CAL_DT',
            'KYLIN_CAL_DT_YEAR_BEG_DT',
            'KYLIN_CAL_DT_QTR_BEG_DT',
            'BUYER_ACCOUNT_ACCOUNT_COUNTRY',
            'BUYER_ACCOUNT_ACCOUNT_BUYER_LEVEL',
            'BUYER_COUNTRY_NAME',
            'BUYER_COUNTRY_COUNTRY',
            'SELLER_ACCOUNT_ACCOUNT_SELLER_LEVEL',
            'SELLER_ACCOUNT_ACCOUNT_COUNTRY',
            'SELLER_COUNTRY_NAME',
            'SELLER_COUNTRY_COUNTRY',
            'KYLIN_CATEGORY_GROUPINGS_CATEG_LVL3_NAME',
            'KYLIN_CATEGORY_GROUPINGS_CATEG_LVL2_NAME',
            'KYLIN_CATEGORY_GROUPINGS_LEAF_CATEG_NAME',
            'KYLIN_CATEGORY_GROUPINGS_META_CATEG_NAME',
        ]
        assert [d.column.name for d in model.dimensions] == [
            'PART_DT',
            'LSTG_FORMAT_NAME',
            'OPS_REGION',
            'MONTH_BEG_DT',
            'CAL_DT',
            'YEAR_BEG_DT',
            'QTR_BEG_DT',
            'ACCOUNT_COUNTRY',
            'ACCOUNT_BUYER_LEVEL',
            'NAME',
            'COUNTRY',
            'ACCOUNT_SELLER_LEVEL',
            'ACCOUNT_COUNTRY',
            'NAME',
            'COUNTRY',
            'CATEG_LVL3_NAME',
            'CATEG_LVL2_NAME',
            'LEAF_CATEG_NAME',
            'META_CATEG_NAME',
        ]
        assert [d.column.alias for d in model.dimensions] == [
            'KYLIN_SALES_PART_DT',
            'KYLIN_SALES_LSTG_FORMAT_NAME',
            'KYLIN_SALES_OPS_REGION',
            'KYLIN_CAL_DT_MONTH_BEG_DT',
            'KYLIN_CAL_DT_CAL_DT',
            'KYLIN_CAL_DT_YEAR_BEG_DT',
            'KYLIN_CAL_DT_QTR_BEG_DT',
            'BUYER_ACCOUNT_ACCOUNT_COUNTRY',
            'BUYER_ACCOUNT_ACCOUNT_BUYER_LEVEL',
            'BUYER_COUNTRY_NAME',
            'BUYER_COUNTRY_COUNTRY',
            'SELLER_ACCOUNT_ACCOUNT_SELLER_LEVEL',
            'SELLER_ACCOUNT_ACCOUNT_COUNTRY',
            'SELLER_COUNTRY_NAME',
            'SELLER_COUNTRY_COUNTRY',
            'KYLIN_CATEGORY_GROUPINGS_CATEG_LVL3_NAME',
            'KYLIN_CATEGORY_GROUPINGS_CATEG_LVL2_NAME',
            'KYLIN_CATEGORY_GROUPINGS_LEAF_CATEG_NAME',
            'KYLIN_CATEGORY_GROUPINGS_META_CATEG_NAME',
        ]
        assert [d.column.datatype for d in model.dimensions] == [
            'DATE',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'DATE',
            'DATE',
            'DATE',
            'DATE',
            'VARCHAR(4096)',
            'INTEGER',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'INTEGER',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
            'VARCHAR(4096)',
        ]

        assert [d.id for d in model.dimensions] == [
            1,
            9,
            11,
            28,
            53,
            84,
            110,
            114,
            116,
            117,
            118,
            122,
            123,
            126,
            127,
            133,
            146,
            149,
            160,
        ]

        assert [d.status for d in model.dimensions] == [
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
            'DIMENSION',
        ]

        assert [m.name for m in model.measures] == [
            'KYLIN_SALES_TRANS_ID_COUNT',
            'KYLIN_SALES_PRICE_SUM',
            'KYLIN_SALES_PRICE_MAX',
            'KYLIN_SALES_ITEM_COUNT_SUM',
            'BUYER_ID_COUNT_DISTINCT',
            'COUNT_ALL',
        ]
        assert [m.verbose for m in model.measures] == [
            'KYLIN_SALES_TRANS_ID_COUNT',
            'KYLIN_SALES_PRICE_SUM',
            'KYLIN_SALES_PRICE_MAX',
            'KYLIN_SALES_ITEM_COUNT_SUM',
            'BUYER_ID_COUNT_DISTINCT',
            'COUNT_ALL',
        ]
        assert [m.measure_type for m in model.measures] == [
            'COUNT',
            'SUM',
            'MAX',
            'SUM',
            'COUNT_DISTINCT',
            'COUNT',
        ]
        assert [m.expression for m in model.measures] == [
            'COUNT (KYLIN_SALES.TRANS_ID)',
            'SUM (KYLIN_SALES.PRICE)',
            'MAX (KYLIN_SALES.PRICE)',
            'SUM (KYLIN_SALES.ITEM_COUNT)',
            'COUNT (DISTINCT KYLIN_SALES.BUYER_ID)',
            'COUNT (1)',
        ]
        assert str(model.from_clause) == (
            '"DEFAULT"."KYLIN_SALES" AS "KYLIN_SALES" '
            'JOIN "DEFAULT"."KYLIN_CAL_DT" AS "KYLIN_CAL_DT" ON "KYLIN_SALES"."PART_DT" = "KYLIN_CAL_DT"."CAL_DT" '  # noqa
            'JOIN "DEFAULT"."KYLIN_ACCOUNT" AS "BUYER_ACCOUNT" ON "KYLIN_SALES"."SELLER_ID" = "BUYER_ACCOUNT"."ACCOUNT_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_ACCOUNT" AS "SELLER_ACCOUNT" ON "KYLIN_SALES"."SELLER_ID" = "SELLER_ACCOUNT"."ACCOUNT_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_CATEGORY_GROUPINGS" AS "KYLIN_CATEGORY_GROUPINGS" ON "KYLIN_SALES"."LEAF_CATEG_ID" = "KYLIN_CATEGORY_GROUPINGS"."LEAF_CATEG_ID" AND "KYLIN_SALES"."LSTG_SITE_ID" = "KYLIN_CATEGORY_GROUPINGS"."SITE_ID" '  # noqa
            'JOIN "DEFAULT"."KYLIN_COUNTRY" AS "BUYER_COUNTRY" ON "BUYER_ACCOUNT"."ACCOUNT_COUNTRY" = "BUYER_COUNTRY"."COUNTRY" '  # noqa
            'JOIN "DEFAULT"."KYLIN_COUNTRY" AS "SELLER_COUNTRY" ON "SELLER_ACCOUNT"."ACCOUNT_COUNTRY" = "SELLER_COUNTRY"."COUNTRY"'  # noqa
        )

        assert model.list_segment()[0].get('id') == '6dcc77dd-8e9b-488a-a52e-ed7cb0245d79'
        assert model.delete('id') == {'success': 'success'}
        assert model.fullbuild() == {'build': 'success'}
        assert model.build(datetime(2000, 1, 1), datetime(2000, 1, 2)) == {'build': 'success'}
        assert model.merge(['123', '234']) == {'build': 'success'}
        assert model.refresh(['123', '234']) == {'build': 'success'}

    def test_list_indexes(self, v4_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v4')
        model = project.get_datasource('kylin_sales_model')
        assert model.list_indexes() == [1, 2, 3, 4]

    def test_build_indexes(self, v4_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v4')
        model = project.get_datasource('kylin_sales_model')
        assert model.build_indexes() == {'build_indexes': 'success'}

    def test_delete_index(self, v4_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v4')
        model = project.get_datasource('kylin_sales_model')
        assert model.delete_index(1) == {'delete_index': 'success'}

    def test_list_index_rules(self, v4_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v4')
        model = project.get_datasource('kylin_sales_model')
        assert model.list_index_rules() == [1, 2, 3, 4]

    def test_clear_up_index_rules(self, v4_api):
        project = create_kylin('kylin://username:password@example/kylin_sales?version=v4')
        model = project.get_datasource('kylin_sales_model')
        assert model.clear_up_index_rules() == {'put_index_rules': 'success'}



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class KylinError(Exception):
    """ Base Exception for Kylin Error """


class KylinUnauthorizedError(KylinError):
    """ Raise when unauthorized error"""


class KylinUserDisabled(KylinError):
    """ Raise when user is disabled """


class NoSuchTableError(KylinError):
    """ Raise when no such table """


class KylinQueryError(KylinError):
    """ Raise when Kylin query error """


class KylinUnsupportedType(KylinError):
    """ Raise when unsupport type in Kylin """


class KylinCubeError(KylinError):
    """ Raise when cube relative operation error """


class KylinJobError(KylinError):
    """ Raise when job relative error """


class KylinModelError(KylinError):
    """ Raise when model relative operation error """


class UnsupportApiError(KylinError):
    """ Raise when use an un-support function"""



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals


class SourceInterface(object):
    source_type = None

    @property
    def name(self):
        raise NotImplementedError()


class ColumnInterface(object):
    @property
    def name(self):
        raise NotImplementedError()

    @property
    def datatype(self):
        raise NotImplementedError()


class DimensionInterface(object):
    @property
    def name(self):
        """ name for SQL query (unique)"""
        raise NotImplementedError()

    @property
    def verbose(self):
        """ verbose for identity dimension (unique)"""
        raise NotImplementedError()

    @property
    def datatype(self):
        raise NotImplementedError()


class MeasureInterface(object):
    @property
    def name(self):
        """ name for SQL query as clause (unique)"""
        raise NotImplementedError()

    @property
    def verbose(self):
        """ verbose for identity measure (unique)"""
        raise NotImplementedError()

    @property
    def measure_type(self):
        """ type of measure, eg: SUM, COUNT """
        raise NotImplementedError()

    @property
    def expression(self):
        """ expression for SQL aggregation"""
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from datetime import date, datetime

import pytest

from kylinpy.exceptions import KylinUnsupportedType
from kylinpy.utils.kylin_types import kylin_to_python


def test_kylin_types():
    with pytest.raises(KylinUnsupportedType):
        raise kylin_to_python('UnsupportedType', 'abc')

    assert kylin_to_python('varchar', 'abc') == 'abc'
    assert kylin_to_python('varchar(256)', 'abc') == 'abc'
    assert kylin_to_python('DECIMAL(20, 6)', '3.1415926') == 3.1415926
    assert kylin_to_python('BIGINT not null', '123456') == 123456
    assert kylin_to_python('INTEGER null', '123456') == 123456
    assert kylin_to_python('DECIMAL(20, 6) NOT null', '123456.123') == 123456.123

    assert kylin_to_python('CHAR', 'abc') == 'abc'
    assert kylin_to_python('VARCHAR', 'abc') == 'abc'
    assert kylin_to_python('STRING', 'abc') == 'abc'
    assert kylin_to_python('DECIMAL', '3.1415926') == 3.1415926
    assert kylin_to_python('DOUBLE', '3.1415926') == 3.1415926
    assert kylin_to_python('FLOAT', '3.1415926') == 3.1415926
    assert kylin_to_python('BIGINT', '3') == 3
    assert kylin_to_python('LONG', '3') == 3
    assert kylin_to_python('INTEGER', '3') == 3
    assert kylin_to_python('INT', '3') == 3
    assert kylin_to_python('TINYINT', '3') == 3
    assert kylin_to_python('SMALLINT', '3') == 3
    assert kylin_to_python('INT4', '3') == 3
    assert kylin_to_python('LONG8', '3') == 3
    assert kylin_to_python('BOOLEAN', 'true') is True
    assert kylin_to_python('BOOLEAN', 'false') is False
    assert kylin_to_python('DATE', '2001-03-10') == date(2001, 3, 10)
    assert kylin_to_python('DATETIME', '2001-03-10 12:12:12') \
        == datetime(2001, 3, 10, 12, 12, 12)
    assert kylin_to_python('TIMESTAMP', '2001-03-10 12:12:12.000000') \
        == datetime(2001, 3, 10, 12, 12, 12)



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from datetime import datetime

from kylinpy.utils.compat import to_millisecond_timestamp, to_second_timestamp


def test_to_millisecond_timestamp():
    dt = datetime(2020, 1, 1)
    dt_1 = datetime(2020, 1, 31, 12, 12, 0)
    dt_2 = datetime(1970, 1, 1)
    assert to_millisecond_timestamp(dt) == 1577836800000
    assert to_millisecond_timestamp(dt_1) == 1580472720000
    assert to_millisecond_timestamp(dt_2) == 0


def test_to_second_timestamp():
    dt = datetime(2020, 1, 1)
    dt_1 = datetime(2020, 1, 31, 12, 12, 0)
    dt_2 = datetime(1970, 1, 1)
    assert to_second_timestamp(dt) == 1577836800
    assert to_second_timestamp(dt_1) == 1580472720
    assert to_second_timestamp(dt_2) == 0



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .cube_source import CubeSource  # noqa
from .table_source import TableSource  # noqa
from .ke4_model_source import KE4ModelSource  # noqa



# -*- coding: utf-8 -*-
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import itertools

import sqlalchemy.exc
from sqlalchemy import pool
from sqlalchemy.engine import default
from sqlalchemy.sql import compiler

from kylinpy.exceptions import NoSuchTableError
from kylinpy.kylindb import Connection
from kylinpy.utils.keywords import CALCITE_KEYWORDS
from kylinpy.utils.sqla_types import kylin_to_sqla

SUPERSET_KEYWORDS = set([
    '__timestamp',
])


class KylinIdentifierPreparer(compiler.IdentifierPreparer):
    compiler.IdentifierPreparer.reserved_words = \
        set(itertools.chain(*[[e.lower(), e] for e in CALCITE_KEYWORDS]))
    compiler.IdentifierPreparer.reserved_words.update(SUPERSET_KEYWORDS)

    def __init__(self, dialect, initial_quote='"',
                 final_quote=None, escape_quote='"', omit_schema=True):
        super(KylinIdentifierPreparer, self).__init__(
            dialect, initial_quote, final_quote, escape_quote, omit_schema,
        )

    def format_label(self, label, name=None):
        return self.quote(name or label.name)


class KylinSQLCompiler(compiler.SQLCompiler):
    _cached_metadata = set()

    def __init__(self, *args, **kwargs):
        super(KylinSQLCompiler, self).__init__(*args, **kwargs)

    def _compose_select_body(self, text, select, inner_columns, froms, byfrom, kwargs):
        text = super(KylinSQLCompiler, self)._compose_select_body(
            text, select, inner_columns, froms, byfrom, kwargs)
        return text

    def visit_column(self, *args, **kwargs):
        result = super(KylinSQLCompiler, self).visit_column(*args, **kwargs)
        return result

    def visit_label(self, *args, **kwargs):
        self.__class__._cached_metadata.add([c.name for c in args][0])
        result = super(KylinSQLCompiler, self).visit_label(*args, **kwargs)
        return result


class KylinDialect(default.DefaultDialect):
    def get_primary_keys(self, connection, table_name, schema=None, **kw):
        pass

    name = 'kylin'
    driver = 'kylin'

    statement_compiler = KylinSQLCompiler
    preparer = KylinIdentifierPreparer

    preexecute_pk_sequences = True
    supports_pk_autoincrement = True
    supports_sequences = True
    sequences_optional = True
    supports_native_decimal = True
    supports_default_values = True
    supports_native_boolean = True
    poolclass = pool.SingletonThreadPool
    supports_unicode_statements = True

    default_paramstyle = 'pyformat'

    def __init__(self, *args, **kwargs):
        super(KylinDialect, self).__init__(*args, **kwargs)

    @classmethod
    def dbapi(cls):
        return Connection

    def initialize(self, connection):
        self.server_version_info = None
        self.default_schema_name = None
        self.default_isolation_level = None
        self.returns_unicode_strings = True

    def create_connect_args(self, url):
        kwargs = {
            'host': url.host,
            'port': url.port or 7070,
            'username': url.username,
            'password': url.password,
            'project': url.database or 'default',
        }
        kwargs.update(url.query)
        if 'is_ssl' in kwargs:
            if kwargs['is_ssl'] in ['1', 'true', 'True', True]:
                kwargs.update({'is_ssl': True})
            else:
                kwargs.update({'is_ssl': False})
        return [[], kwargs]

    def do_execute(self, cursor, statement, parameters, context=None):
        super(KylinDialect, self).do_execute(cursor, statement, parameters, context)

    def get_table_names(self, connection, schema=None, **kw):
        conn = connection.connect()
        tables = conn.connection.connection.get_all_tables(schema)
        return tables

    def get_schema_names(self, connection, schema=None, **kw):
        conn = connection.connect()
        schemas = conn.connection.connection.get_all_schemas()
        return schemas

    def has_table(self, connection, table_name, schema=None):
        # disable check table exists
        return False

    def has_sequence(self, connection, sequence_name, schema=None):
        return False

    def get_columns(self, connection, table_name, schema=None, **kw):
        conn = connection.connect()
        try:
            columns = conn.connection.connection.get_table_source(table_name, schema).columns
            return [{
                'name': col.name,
                'type': kylin_to_sqla(col.datatype),
            } for col in columns]
        except NoSuchTableError:
            raise sqlalchemy.exc.NoSuchTableError

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        return []

    def get_indexes(self, connection, table_name, schema=None, **kw):
        return []

    def get_view_names(self, connection, schema=None, **kw):
        return []

    def get_pk_constraint(self, conn, table_name, schema=None, **kw):
        return {}

    def get_unique_constraints(self, connection, table_name, schema=None, **kw):
        return []
